@@ -1,4 +1,5 @@
 from airy.core.conf import settings
+from airy.core.exceptions import Http404
 from airy.core.files.uploadedfile import SimpleUploadedFile
 from tornado.web import *
 from tornado.escape import *
@@ -16,7 +17,7 @@ class AirySite(object):
     application = None
     loader = None
 
-    def resolve_url(self, url, connection, arguments=[]):
+    def resolve_url(self, url, connection, arguments={}):
         handler = None
         args = []
         kwargs = {}
@@ -130,6 +131,13 @@ class AiryRequestHandler(RequestHandler):
         template_args.update(kwargs)
         return super(AiryRequestHandler, self).render_string(template_name, **template_args)
 
+    def is_robot(self, *args, **kwargs):
+        robots = ['Googlebot', ]
+        robots.extend(args)
+        for robot in robots:
+            if robot in self.request.headers['User-Agent']:
+                return True
+        return False
 
 
 class AiryHandler(object):
@@ -189,7 +197,8 @@ class AiryHandler(object):
                         self.files[k] = SimpleUploadedFile(v['name'], file_content, mimetype)
                         is_file_content = True
                     except Exception, e:
-                        print "Failed to decode argument '%s'" % k
+                        logging.error("Failed to decode argument '%s'" % k)
+                        continue
             if not is_file_content:
                 self.arguments[k] = values
         for k,v in kwargs.iteritems():
@@ -564,16 +573,20 @@ class AiryCoreHandler(SocketConnection):
         (scheme, host, path, arguments) = self.parse_url(url)
         self.set_state(path)
         handler, hargs, hkwargs = site.resolve_url(path, self, arguments=arguments)
+        hargs.extend(args)
+        hkwargs.update(kwargs)
         if handler:
             logging.info('GET %s from %s %s => %s' % (path, self.info.ip,
                                                       self.info.arguments.get('t', ''),
                                                       handler.__class__.__name__))
-            hargs.extend(args)
-            hkwargs.update(kwargs)
-            handler.get(*hargs, **hkwargs)
+            try:
+                handler.get(*hargs, **hkwargs)
+            except Http404:
+                self.process_404(*hargs, **hkwargs)
         else:
             logging.error('GET %s from %s %s failed: no handler found' % \
                           (path, self.info.ip, self.info.arguments.get('t', '')))
+            self.process_404(*hargs, **hkwargs)
 
     @event('post')
     def post(self, url, *args, **kwargs):
@@ -590,21 +603,75 @@ class AiryCoreHandler(SocketConnection):
         arguments.update(data)
         self.set_state(path)
         handler, hargs, hkwargs = site.resolve_url(path, self, arguments=arguments)
+        hargs.extend(args)
+        hkwargs.update(kwargs)
         if handler:
             logging.info('POST %s from %s %s => %s' % (path, self.info.ip,
                                                       self.info.arguments.get('t', ''),
                                                       handler.__class__.__name__))
-            hargs.extend(args)
-            hkwargs.update(kwargs)
-            handler.post(*hargs, **hkwargs)
+            try:
+                handler.post(*hargs, **hkwargs)
+            except Http404:
+                self.process_404(*hargs, **hkwargs)
         else:
             logging.error('POST %s from %s %s failed: no handler found' % \
                           (path, self.info.ip, self.info.arguments.get('t', '')))
+            self.process_404(*hargs, **hkwargs)
 
     def on_close(self):
         logging.info('Socket Disconnected: %s %s' % (self.info.ip, self.info.arguments.get('t', '')))
         site.connections.remove(self)
 
+    def process_404(self, *args, **kwargs):
+        try:
+            handler_path = settings.page_not_found_handler
+            path, name = handler_path.rsplit('.', 1)
+            handler_module = __import__(path, fromlist=path)
+            handler_class = getattr(handler_module, name)
+            handler_class(site, self).get()
+        except AttributeError:
+            logging.error('Page with url %s doesn\'t exist' % (kwargs.get('url', ''), ))
+
+    #
+    # Airy client interaction
+    #
+    def execute(self, data):
+        "Execute the given data on the client side"
+        self.emit('execute', data)
+        return self
+
+    def redirect(self, url):
+        return self.execute('airy.request("get", "%s");' % url)
+
+    def insert(self, target, data):
+        "Insert data into target"
+        return self.execute('airy.ui.insert("%s", %s);' % (target, json_encode(data)))
+
+    def append(self, target, data):
+        "Append data into target"
+        return self.execute('airy.ui.append("%s", %s);' % (target, json_encode(data)))
+
+    def prepend(self, target, data):
+        return self.execute('airy.ui.prepend("%s", %s);' % (target, json_encode(data)))
+
+    def render_string(self, template_name, **kwargs):
+        "Render the given template"
+        context_processors = getattr(settings, 'template_context_processors', [])
+        template_args = {'reverse_url': self.reverse_url}
+        for processor_path in context_processors:
+            path, name = processor_path.rsplit('.', 1)
+            processor_module = __import__(path, fromlist=path)
+            processor = getattr(processor_module, name)
+            template_args.update(processor(self, **kwargs))
+        template_args.update(kwargs)
+        html = self.site.loader.load(template_name).generate(**template_args)
+        return html
+
+    def render(self, target, template_name, **kwargs):
+        "Render into the target element (jQuery selector)"
+        html = self.render_string(template_name, **kwargs)
+        self.insert(target, html)
+        return self
 
 
 core_router = TornadioRouter(AiryCoreHandler)
